from ..dataModels.models import Port
from ..dataModels.models import Service
from .fileParser import fileParser
import os

class WriterError(Exception):

    """Base class for exceptions in the writer module."""

    def __init__(self, message, filename=None):
        self.message = message
        self.filename = filename
        super().__init__(message)



class fileWriter:

    """
    Class for writing or updating .qpmgr files
    """

    def __init__(self, filename, autoFlush = True, ignoreExtension = False):

        """
        Constructor for fileWriter class

        Args:
            filename (str): Path to the file
            autoFlush (bool): Whether to flush data to file or not
            ignoreExtension (bool): Whether to ignore file extension or not
        """

        self.__portNumError = WriterError('Port should be number', filename)

        # Variables; consts
        self.__autoFlush = autoFlush
        if ignoreExtension: self.__filename = filename
        else: self.__filename = filename if filename.endswith(".qpmgr") else f"{filename}.qpmgr" # Shadows filename into class scope
        self.__servicesList = {} # Parsed information from .qpmgr file w/service as key
        self.__stateTable = {"O" : True, "C" : False}
        self.__serviceRecordMask = "$SERVICE -- $PORTSINFO || $SERVICEDESCRIPTION"
        self.__fileHeader = \
"""\
#====================== firewall allowed ports ======================#
#                                                                    #
#  |--------------------------------------------------------------|  #
#  |                             HINT                             |  #
#  |                                                              |  #
#  |   Use pManager to work with .qpmgr files or follow pattern   |  #
#  |--------------------------------------------------------------|  #
#  |                    ScriptReadable Pattern                    |  #
#  |                                                              |  #
#  | Service -- TCP/UDP::port[1-N]::O/C::PortDescr || ServiceDesc |  #
#  |--------------------------------------------------------------|  #
#                                                                    #
#====================================================================#

"""
        if os.path.exists(self.__filename):
            fp = fileParser(self.__filename)
            existingServicesList = fp.getPlainServiceList()
            for _ in existingServicesList.keys():
                self.constructServiceObj(_, existingServicesList[_]['description'], existingServicesList[_]['ports'])

    def __flushFile(self):

        """
        Flushes data to file
        """

        with open(self.__filename, 'w') as file:

            file.write(self.__fileHeader)

            for srv in self.__servicesList.keys():
                file.write(self.__serviceRecordMask.replace('$SERVICE', srv, 1)
                           .replace('$PORTSINFO', self.__servicesList[srv].generatePortConfig(), 1)
                           .replace('$SERVICEDESCRIPTION', self.__servicesList[srv].getServiceDescription(), 1) + '\n'
                           )

    def constructServiceObj(self, serviceName, serviceDescription ="No description", portList=None) -> Service:

        """
        Creates service object and adds it to the list

        Args:
            serviceName (str): Name of the service
            serviceDescription (str): Description of the service
            portList (dict): Dictionary of ports, where key is port number and value is port object

        Returns:
            Service (pManager.dataModels.models.Service): Created service object
        """

        srv = Service(serviceName=serviceName, serviceDescription=serviceDescription, portList=portList)
        self.addService(srv)
        return srv

    def bindPortToService(self, serviceName, portNum, portProto, openState, portDesc ="No description") -> Port:

        """
        Binds port to existing service

        Args:
            serviceName (str): Name (key) of the service
            portNum (int): Port number
            portProto (str): Protocol of the port
            openState (bool): State of the port
            portDesc (str): Description of the port

        Returns:
            Port (pManager.dataModels.models.Port): Port object that was created
        """

        port = Port(serviceName, portNum, portProto, openState, portDesc)
        self.addPort(port)
        return Port(serviceName, portNum, portProto, openState, portDesc)

    def addService(self, serviceObj: Service):

        """
        Adds service to the list

        Args:
            serviceObj (pManager.dataModels.models.Service): Service object

        Raises:
            WriterError: If the service object is not of type Service
        """

        if type(serviceObj) == Service:
            self.__servicesList[serviceObj.getServiceName()] = serviceObj
        else:
            raise WriterError("Wrong type of object passed to addService method", self.__filename)

    def addPort(self, portObj: Port):

        """
        Adds port to existing service

        Args:
            portObj (pManager.dataModels.models.Port): Port object

        Raises:
            WriterError: If the port object is not of type Port
        """

        if type(portObj) == Port:

            if self.__servicesList[portObj.getServiceName()].isPortsAdded:
                self.__servicesList[portObj.getServiceName()].addPortToService(portObj.getPortAsJson())
            else:
                self.__servicesList[portObj.getServiceName()].addPortToService(portObj.getPortAsJson())

            if self.__autoFlush: self.__flushFile()

        else:

            raise WriterError("Wrong type of object passed to addPort method", self.__filename)

    def getServiceList(self):

        """
        Returns dictionary of services

        Returns:
            dict: Dictionary of services
        """

        return self.__servicesList

    def writeToFile(self):

        """
        Writes service list to file
        """

        self.__flushFile()


# Intializing errors

class ParserError(Exception):
    """Base class for exceptions in the parser module."""
    def __init__(self, message, filename=None):
        self.message = message
        self.filename = filename
        super().__init__(message)

class ConfigurationError(ParserError):
    """Exception raised for configuration-related errors during file parsing."""
    pass

class UnknownServiceError(ParserError):
    """Exception raised when an unknown service is encountered during parsing."""
    pass

class UnknownPortError(ParserError):
    """Exception raised when an unknown port is encountered during parsing."""
    pass

class FormatError(ParserError):
    """Exception raised for errors in the file format."""
    pass



class fileParser:

    """
    Class for parsing .qpmgr file
    """

    def __init__(self, filename, ignoreExtension = False):

        """
        Constructor for fileParser class

        Args:
            filename (str): Path to the file
            ignoreExtension (bool): Whether to ignore file extension or not
        """

        # Variables
        if ignoreExtension: self.__filename = filename
        else: self.__filename = filename if filename.endswith(".qpmgr") else f"{filename}.qpmgr"
        self.__servicesList = {} # Parsed information from .qpmgr file w/service as key
        self.__portList = {} # Parsed information from .qpmgr file w/port as key, w/o service description
        self.__openedPorts = []
        self.__closedPorts = []
        self.__blankLines = ("\n", "", " ") # What lines parser should skip in file?
        self.__stateTable = {"O": True, "C": False}
        # Port state: O = Opened = True
        #             C = Closed = False

        # Errors declare
        self.__filenameError = ConfigurationError("Filename isn't set. You can get filename from errorObj.filename", self.__filename)
        self.__fileNotFoundError = ParserError(f"File {self.__filename} not found. ", self.__filename)
        self.__unknownServiceError = UnknownServiceError("Unknown service", self.__filename)
        self.__unknownPortError = UnknownPortError("Unknown port", self.__filename)
        self.__formatError = ParserError("File is blank or not formated as .qpmgr", self.__filename)

        # Starting fileParser with configuration above
        self.__parseFile()

    def __parseFile(self):
        if self.__filename is None: raise self.__filenameError

        try:
            with open(self.__filename, 'r') as file:
                for line in file.readlines():

                    if line.startswith("#"): continue
                    if line in self.__blankLines: continue

                    line = line.replace("\n", "")
                    data = line.split(' || ')[0]
                    serviceName = data.split(' -- ')[0]
                    ports = {}

                    for port in data.split(' -- ')[1].split('; '):
                        port = port.split('::')
                        ports[int(port[1])] = {"protocol": port[0], "description": port[3], "opened": self.__stateTable[port[2]]}

                    desc = line.split(' || ')[1]

                    self.__servicesList[serviceName] = {"ports": ports, "description": desc}

                file.close()

        except FileNotFoundError: raise self.__fileNotFoundError
        except IndexError: raise self.__formatError

        for srv in self.__servicesList.keys():
            for port in self.__servicesList[srv]['ports']:
                 self.__portList[port] = {
                     "serviceName": srv,
                     "protocol": self.__servicesList[srv]['ports'][port]['protocol'],
                     "description": self.__servicesList[srv]['ports'][port]['description'],
                     "opened": self.__servicesList[srv]['ports'][port]['opened']
                 }

        self.__fetchPorts()

        if self.__servicesList == {}: raise

    def __fetchPorts(self):
        if self.__servicesList == {}: self.__parseFile()
        for service in self.__servicesList:
            for port in self.__servicesList[service]['ports'].keys():
                if self.__servicesList[service]['ports'][port]['opened']: self.__openedPorts.append(port)
                else: self.__closedPorts.append(port)

    def getOpenedPorts(self) -> list:

        """
        Returns list of opened ports

        Returns:
            list: List of opened ports
        """

        if (self.__openedPorts == {}) or (self.__closedPorts == {}): self.__fetchPorts()
        return self.__openedPorts

    def getClosedPorts(self) -> list:

        """
        Returns list of closed ports

        Returns:
            list: List of closed ports
        """

        if (self.__openedPorts == {}) or (self.__closedPorts == {}): self.__fetchPorts()
        return self.__closedPorts

    def getServicesList(self) -> list:

        """
        Returns list of services

        Returns:
            list: List of services
        """

        if self.__servicesList == {}: self.__parseFile()
        return list(self.__servicesList.keys())

    def getOpenedPortsByService(self, serviceName) -> list:

        """
        Returns list of opened ports by service

        Args:
            serviceName (str): Name of the service

        Returns:
            list: List of opened ports
        """

        if self.__servicesList == {}: self.__parseFile()
        serviceInfo = self.__servicesList.get(serviceName)
        if serviceInfo is None: raise self.__unknownServiceError
        return list(map(str, serviceInfo['ports'].keys()))

    def getDescriptionByService(self, serviceName) -> str:

        """
        Returns description of the service by it\'s name

        Args:
            serviceName (str): Name of the service

        Returns:
            str: Description of the service
        """

        if self.__servicesList == {}: self.__parseFile()
        serviceInfo = self.__servicesList.get(serviceName)
        if serviceInfo is None: raise self.__unknownServiceError
        return serviceInfo['description']

    def getBeautifuledInfoByService(self, serviceName) -> str:


        """
        Returns string with beautifuled information about the service by it\'s name

        Args:
            serviceName (str): Name of the service

        Returns:
            str: String with beautifuled information about the service
        """

        if self.__servicesList == {}: self.__parseFile()
        return \
f'''\
Service name: {serviceName}
Ports in use: {', '.join(self.getOpenedPortsByService(serviceName))}
Description: {self.getDescriptionByService(serviceName)}\
'''

    def getServiceByPort(self, portNum) -> str:

        """
        Returns service name by port number

        Args:
            portNum (int): Number of the port

        Returns:
            str: Service name
        """

        if self.__portList == {}: self.__parseFile()
        portInfo = self.__portList.get(portNum)
        if portInfo is None: raise self.__unknownPortError
        return portInfo['serviceName']

    def getDescriptionByPort(self, portNum) -> str:

        """
        Returns description of the port by its number

        Args:
            portNum (int): Number of the port

        Returns:
            str: Description of the port
        """

        if self.__servicesList == {}: self.__parseFile()
        portInfo = self.__portList.get(portNum)
        if portInfo is None: raise self.__unknownPortError
        return portInfo['description']

    def getProtocolByPort(self, portNum) -> str:

        """
        Returns protocol of the port by its number

        Args:
            portNum (int): Number of the port

        Returns:
            str: Protocol of the port
        """

        if self.__portList == {}: self.__parseFile()
        portInfo = self.__portList.get(portNum)
        if portInfo is None: raise self.__unknownPortError
        return portInfo['protocol']

    def getPlainServiceList(self) -> dict:

        """
        Returns dictionary with plain service list

        Returns:
            dict: Plain service list
        """

        if self.__servicesList == {}: self.__parseFile()
        return self.__servicesList

    def getPlainPortList(self) -> dict:

        """
        Returns dictionary with plain port list

        Returns:
            dict: Plain port list
        """

        if self.__portList == {}: self.__parseFile()
        return self.__portList


class Port:

    """
    Class for representing a port in .qpmgr file.
    """

    def __init__(self, serviceName, portNum, portProto, openState, portDesc = "No description"):

        # Port stateList
        self.__portOpenState = {"O": True, "C": False}

        # Variables
        self.__serviceName = serviceName
        self.__portNum = int(portNum)
        self.__portProtocol = portProto
        self.__portDescription = portDesc
        if type(openState) == bool: self.__opened = openState
        else: self.__opened = self.__portOpenState[openState]

    def getPortAsJson(self):

        """
        Returns the port details as a JSON-like dictionary.

        Returns:
            dict: A dictionary where the key is the port number, and the value is another dictionary containing:
                - "protocol" (str): The protocol of the port.
                - "description" (str): The description of the port.
                - "opened" (bool): The state of the port, whether it is open or closed.
        """

        return {
            self.__portNum: {
                "protocol": self.__portProtocol,
                "description": self.__portDescription,
                "opened": self.__opened
            }
        }

    def getServiceName(self):

        """
        Returns the name of the service associated with the port.

        Returns:
            str: The name of the service.
        """

        return self.__serviceName


class Service:

    """
    Class for representing a service in .qpmgr file.
    """

    def __init__(self, serviceName, portList=None, serviceDescription ="No description"):

        if portList is None: portList = {
            -1: {
                "protocol": "TCP",
                "description": "PORTLIST NOT SET",
                "opened": False
            }
        }

        self.isPortsAdded = False
        self.__portList = portList
        self.__serviceName = serviceName
        self.__serviceDescription = serviceDescription
        self.__portRecordMask = "$PROTOCOL::$PORT::$STATE::$DESCRIPTION"
        print(f'portList: {self.__portList}\ndescription: {self.__serviceDescription}')

    def addPortToService(self, portJson):

        """
        Adds a port to the service.

        Args:
            portJson (dict): Json of the port to add
        """

        if self.isPortsAdded:
            self.__portList.update(portJson)
        else:
            self.__portList = portJson
            self.isPortsAdded = True


    def getServiceName(self):

        """
        Returns the name of the service.

        Returns:
            str: The name of the service.
        """

        return self.__serviceName

    def getServiceDescription(self):

        """
        Returns the description of the service.

        Returns:
            str: The description of the service.
        """

        return self.__serviceDescription

    def generatePortConfig(self):

        """
        Returns the port configuration of the service.

        Returns:
            str: The port configuration of the service.
        """

        ports = []
        if type(self.__portList) is dict:
            for __p in self.__portList.keys():
                port = self.__portList[__p]
                portConfig = (self.__portRecordMask
                              .replace("$PROTOCOL", port['protocol'], 1)
                              .replace("$PORT", str(__p), 1)
                              .replace("$STATE", "O" if port['opened'] else "C", 1)
                              .replace("$DESCRIPTION", port['description'], 1)
                              )
                ports.append(portConfig)
        else: ports = ["-1::TCP::C::PORTLIST NOT SET"]
        return "; ".join(ports)

    def getServiceAsJson(self):

        """
        Returns the service details as a JSON-like dictionary.

        Returns:
            dict: A dictionary where the key is the service name, and the value is another dictionary containing:
                - "ports" (dict): A dictionary where the key is the port number, and the value is another dictionary containing:
                    - "protocol" (str): The protocol of the port.
                    - "description" (str): The description of the port.
                    - "opened" (bool): The state of the port, whether it is open or closed.
                - "description" (str): The description of the service.
        """

        return {
            self.__serviceName: {
                "ports": self.__portList,
                "description": self.__serviceDescription
            }
        }


from setuptools import setup
from pathlib import Path
longDescription = (Path(__file__).parent / "README.md").read_text()

setup(
    name='portsManager',
    version='0.1.4',
    packages=['pManager', 'pManager.modules', 'pManager.dataModels'],
    url='https://github.com/Qwantman/pManager',
    license='GPL-3.0',
    author='Qwantman',
    author_email='svr0116@gmail.com',
    description='Library to work with .qpmgr files',
    long_description=longDescription,
    long_description_content_type='text/markdown',
    python_requires='>=3.10',
    programming_language='python3'
)
